Module reference

Module reference 

๐Ÿ‘ŽDeprecated since 0.17.1:

Use Ref from the crate root instead.

Structs

Ref
A reference to a value in a link section. This allows platforms like WASM to reference the value, even though the final location is not known until after initialization.