pub struct Ref<T: 'static> { /* private fields */ }Expand description
A reference to a value in a link section. This allows platforms like WASM to reference the value, even though the final location is not known until after initialization.
Implementations