pub struct MovableBackref<T: 'static> { /* private fields */ }Expand description
A backref record submitted alongside each item in a TypedMovableSection.
This points to a MovableRef that lives outside of the section.
Implementations
Sourceimpl<T> MovableBackref<T>
impl<T> MovableBackref<T>
Sourcepub fn current_ptr(&self) -> *const T
pub fn current_ptr(&self) -> *const T
Current value of the stable pointer slot as a pointer.
Sourcepub unsafe fn set_current_ptr(&self, ptr: *const T)
pub unsafe fn set_current_ptr(&self, ptr: *const T)
Update the stable pointer slot.
Safety
Updating the slot requires exclusive access. See
Exclusive access for more information. Any live
&T or MovableRef dereference may alias the old or new target.